
/********************************************
jmp.c
copyright 1991, Michael D. Brennan

This is a source file for mawk, an implementation of
the AWK programming language.

Mawk is distributed without warranty under the terms of
the GNU General Public License, version 2, 1991.
********************************************/

/* $Log: jmp.c,v $
 * Revision 1.4  1995/06/18  19:42:19  mike
 * Remove some redundant declarations and add some prototypes
 *
 * Revision 1.3  1995/04/21  14:20:16  mike
 * move_level variable to fix bug in arglist patching of moved code.
 *
 * Revision 1.2  1993/07/14  13:17:49  mike
 * rm SIZE_T and run thru indent
 *
 * Revision 1.1.1.1  1993/07/03	 18:58:15  mike
 * move source to cvs
 *
 * Revision 5.3	 1993/01/09  19:03:44  mike
 * code_pop checks if the resolve_list needs relocation
 *
 * Revision 5.2	 1993/01/07  02:50:33  mike
 * relative vs absolute code
 *
 * Revision 5.1	 1991/12/05  07:56:10  brennan
 * 1.1 pre-release
 *
*/

/* this module deals with back patching jumps, breaks and continues,
   and with save and restoring code when we move code.
   There are three stacks.  If we encounter a compile error, the
   stacks are frozen, i.e., we do not attempt error recovery
   on the stacks
*/


#include "mawk.h"
#include "symtype.h"
#include "jmp.h"
#include "code.h"
#include "sizes.h"
#include "init.h"
#include "memory.h"

#define error_state  (MAWK->compile_error_count>0)


/*---------- back patching jumps  ---------------*/

void
mawk_code_jmp(MAWK, jtype, target)
   mawk_state_t *MAWK;
   int jtype ;
   INST *target ;
{
   if (error_state)  return ;

   /* WARNING: Don't emit any code before using target or
     relocation might make it invalid */

   if (target)	mawk_code2op(MAWK, jtype, target - (mawk_code_ptr + 1)) ;
   else
   {
      register JMP *p = ZMALLOC(JMP) ;

      /* stack for back patch */
      mawk_code2op(MAWK, jtype, 0) ;
      p->source_offset = mawk_code_offset - 1 ;
      p->link = MAWK->jmp_top ;
      MAWK->jmp_top = p ;
   }
}

void
mawk_patch_jmp(MAWK, target)		/* patch a jump on the jmp_stack */
   mawk_state_t *MAWK;
   INST *target ;
{
   register JMP *p ;
   register INST *source ;	 /* jmp starts here */

   if (!error_state)
   {
#ifdef	DEBUG
      if (!MAWK->jmp_top)  mawk_bozo(MAWK, "jmp stack underflow") ;
#endif

      p = MAWK->jmp_top ; MAWK->jmp_top = p->link ;
      source = p->source_offset + mawk_code_base ;
      source->op = target - source ;

      ZFREE(p) ;
   }
}


/*-- break and continue -------*/

void
mawk_BC_new(mawk_state_t *MAWK)			/* mark the start of a loop */
{
   mawk_BC_insert(MAWK, 0, (INST *) 0) ;
}

void
mawk_BC_insert(MAWK, type, address)
mawk_state_t *MAWK;
int type ; INST *address ;
{
   register BC *p ;

   if (error_state)  return ;

   if (type && !MAWK->bc_top)
   {
      mawk_compile_error(MAWK, "%s statement outside of loop",
		    type == 'B' ? "break" : "continue") ;

      return ;
   }
   else
   {
      p = ZMALLOC(BC) ;
      p->type = type ;
      p->source_offset = address - mawk_code_base ;
      p->link = MAWK->bc_top ;
      MAWK->bc_top = p ;
   }
}


/* patch all break and continues for one loop */
void
mawk_BC_clear(MAWK, B_address, C_address)
   mawk_state_t *MAWK;
   INST *B_address, *C_address ;
{
   register BC *p, *q ;
   INST *source ;

   if (error_state)  return ;

   p = MAWK->bc_top ;
   /* pop down to the mark node */
   while (p->type)
   {
      source = mawk_code_base + p->source_offset ;
      source->op = (p->type == 'B' ? B_address : C_address)
	 - source ;

      q = p ; p = p->link ; ZFREE(q) ;
   }
   /* remove the mark node */
   MAWK->bc_top = p->link ;
   ZFREE(p) ;
}

/*-----	 moving code --------------------------*/

#define	 NO_SCOPE	-1
 /* means relocation of resolve list not needed */

void
mawk_code_push(MAWK, code, len, scope, fbp)
   mawk_state_t *MAWK;
   INST *code ;
   unsigned len ;
   int scope ;
   FBLOCK *fbp ;
{
   register MC *p ;

   if (!error_state)
   {
      p = ZMALLOC(MC) ;
      p->len = len ;
      p->link = MAWK->mc_top ;
      MAWK->mc_top = p ;

      if (len)
      {
	 p->code = (INST *) zmalloc(MAWK, sizeof(INST) * len) ;
	 memcpy(p->code, code, sizeof(INST) * len) ;
      }
      if (!resolve_list)  p->scope = NO_SCOPE ;
      else
      {
	 p->scope = scope ;
	 p->move_level = MAWK->ps.code_move_level ;
	 p->fbp = fbp ;
	 p->offset = code - mawk_code_base ;
      }
   }
   MAWK->ps.code_move_level++ ;
}

/* copy the code at the top of the mc stack to target.
   return the number of INSTs moved */

unsigned
mawk_code_pop(MAWK, target)
   mawk_state_t *MAWK;
   INST *target ;
{
   register MC *p ;
   unsigned len ;
   int target_offset ;

   if (error_state)  return 0 ;

#ifdef	DEBUG
   if (!MAWK->mc_top)	 mawk_bozo(MAWK, "mc underflow") ;
#endif

   p = MAWK->mc_top ; MAWK->mc_top = p->link ;
   len = p->len ;

   while (target + len >= mawk_code_warn)
   {
      target_offset = target - mawk_code_base ;
      mawk_code_grow(MAWK) ;
      target = mawk_code_base + target_offset ;
   }

   if (len)
   {
      memcpy(target, p->code, len * sizeof(INST)) ;
      zfree(MAWK, p->code, len * sizeof(INST)) ;
   }

   if (p->scope != NO_SCOPE)
   {
      target_offset = target - mawk_code_base ;
      mawk_relocate_resolve_list(p->scope, p->move_level, p->fbp,
			 p->offset, len, target_offset - p->offset) ;
   }

   ZFREE(p) ;
   MAWK->ps.code_move_level-- ;
   return len ;
}
